from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
import os.path

from .Filespec import Filespec
from .FilesetInfo import FilesetInfo
from .PooledFile import PooledFile
from .util import filetimestr, verbose_stderr, debug_log, warning

class SimpleFilesetCache(object):

    def __init__(self, path, deltadir, mapper, attrs, sel):
        #debug_log("SimpleFilesetCache(%s)::__init__)\n" % path)
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs
        self._sel = sel
        self._filespecs = []    # in-memory read cache
        self._info = {}         # indexed by filter string
        self._file = None
        self._filepos = 0
        self._fileinfo = None
        self._deletedFilelist = {}    # paths of deleted files
        self._deletedInfo = FilesetInfo()

    def filelistpath(self, deleted=False):
        if deleted:
            return os.path.join(self._deltadir, "deleted.filelist")
        else:
            return os.path.join(self._path, "filelist")

    def infopath(self, deleted=False):
        if deleted:
            return os.path.join(self._deltadir, "deleted.info")
        else:
            return os.path.join(self._path, "info")

    def select(self, filter=None, includeDeleted=False):
        # first read from in-memory cache, which may be empty, or partial if last file read was interrupted
        #debug_log("SimpleFilesetCache select %s from memory cache\n" % str(filter))
        for filespec in self._filespecs:
            if includeDeleted or filespec.path not in self._deletedFilelist:
                if filter is None or filter.selects(filespec):
                    #debug_log("SimpleFilesetCache read from memory %s\n" % filespec)
                    yield filespec

        # now read from file, unless this is complete
        if self._filepos is not None:
            #debug_log("reading filelist from %s cache at %s\n" % (filetimestr(self._path), self._path))
            #debug_log("SimpleFilesetCache(%s) select %s from file cache %s\n" % (self._path, str(filter), self._path))
            self._deletedFilelist = {}
            filelist = self.filelistpath()
            deletedFilelist = self.filelistpath(deleted=True)
            try:
                if os.path.exists(deletedFilelist):
                    # if deleted filelist is older than cache, remove it
                    if os.stat(deletedFilelist).st_mtime < os.stat(filelist).st_mtime:
                        #debug_log("removing obsolete deleted filelist %s\n" % deletedFilelist)
                        os.remove(deletedFilelist)
                    else:
                        #debug_log("reading deleted filelist %s\n" % deletedFilelist)
                        with open(deletedFilelist, 'r') as f:
                            for line in f:
                                self._deletedFilelist[line.rstrip('\n')] = True
            except IOError:
                warning("can't read deleted filelist %s, ignoring" % deletedFilelist)
            try:
                with PooledFile(filelist, 'r') as f:
                    if self._filepos != 0:
                        f.seek(self._filepos)
                    #debug_log("SimpleFilesetCache(%s) select %s opened file cache as %s at %d\n" % (self._path, filter, f, self._filepos))
                    try:
                        for filespec in Filespec.fromFile(f, self, self._sel):
                            self._filespecs.append(filespec)
                            if includeDeleted or filespec.path not in self._deletedFilelist:
                                if filter is None or filter.selects(filespec):
                                    #debug_log("SimpleFilesetCache read from file %s\n" % filespec)
                                    yield filespec
                    except:
                        # on any error save the filepos
                        self._filepos = f.tell()
                        #debug_log("SimpleFilesetCache(%s) exception, saving filepos at %d\n" % (self._path, self._filepos))
                        raise
                    # reading file is complete
                    self._filepos = None
            except IOError:
                warning("can't read filelist %s, ignoring" % filelist)

    def merge_info(self, acc, filter=None):
        #debug_log("SimpleFilesetCache(%s) merge_info\n" % self._path)
        if filter is None:
            #debug_log("SimpleFilesetCache(%s)::merge_info(None)\n" % self._path)
            if self._fileinfo is None:
                #debug_log("SimpleFilesetCache(%s)::merge_info(None) reading info file\n" % self._path)
                infofile = self.infopath()
                deletedInfofile = self.infopath(deleted=True)
                try:
                    if os.path.exists(deletedInfofile):
                        # if deleted filelist is older than cache, remove it
                        if os.stat(deletedInfofile).st_mtime < os.stat(infofile).st_mtime:
                            #debug_log("removing obsolete deleted infofile %s\n" % deletedInfofile)
                            os.remove(deletedInfofile)
                        else:
                            #debug_log("reading deleted infofile %s\n" % deletedInfofile)
                            with open(deletedInfofile, 'r') as f:
                                self._deletedInfo = FilesetInfo.fromFile(f)
                except IOError:
                    warning("can't read deleted info %s, ignoring" % deletedInfofile)
                    self._deletedInfo = FilesetInfo()
                try:
                    with open(infofile, 'r') as f:
                        self._fileinfo = FilesetInfo.fromFile(f)
                except IOError:
                    warning("can't read info %s, ignoring" % infofile)
                    self._fileinfo = FilesetInfo()
            info = self._fileinfo
        else:
            f = str(filter)
            #debug_log("SimpleFilesetCache(%s)::merge_info(%s)\n" % (self._path, f))
            if f in self._info:
                info = self._info[f]
            else:
                #debug_log("SimpleFilesetCache(%s)::merge_info(%s) scanning\n" % (self._path, f))
                info = FilesetInfo()
                self._info[f] = info
                for filespec in self.select(filter, includeDeleted=True):
                    info.add(1, filespec.size)
        acc.accumulate(info, self._sel)
        acc.decumulate(self._deletedInfo, self._sel)

    def add(self, filespec):
        if self._fileinfo is None:
            self._fileinfo = FilesetInfo()
        self._fileinfo.add(1, filespec.size)
        # don't cache writes, as this would mean caching the whole of the filelist
        if self._file is None:
            #debug_log("SimpleFilesetCache writing file cache at %s\n" % self._path)
            if not os.path.exists(self._path):
                os.makedirs(self._path)
            self._file = PooledFile(self.filelistpath(), 'w')
        filespec.write(self._file)

    def finalize(self):
        """Finalize writing the cache."""
        if self._file is not None:
            self._file.close()
            self._file = None

        # sort the filelist
        with open(self.filelistpath()) as f:
            sorted_filelist = sorted(f, key=Filespec.formattedToPath)
        with open(self.filelistpath(), 'w') as f:
            f.writelines(sorted_filelist)

        # write info file
        with open(self.infopath(), 'w') as infofile:
            if self._fileinfo is not None:
                self._fileinfo.write(infofile)

    def delete(self, filespec):
        #debug_log("SimpleFilesetCache(%s) delete %s\n" % (self._path, filespec.path))
        self._deletedFilelist[filespec.path] = True
        self._deletedInfo.add(1, filespec.size)

    def saveDeletions(self):
        if len(self._deletedFilelist) > 0:
            deletedFilelist = self.filelistpath(deleted=True)
            #debug_log("SimpleFilesetCache(%s)::saveDeletions deletedFilelist\n" % self._path)
            try:
                if not os.path.exists(self._deltadir):
                    os.makedirs(self._deltadir)
                with open(deletedFilelist, 'w') as f:
                    for path in self._deletedFilelist:
                        f.write("%s\n" % path)
            except IOError:
                warning("can't write deleted filelist %s, ignoring" % deletedFilelist)

        if self._deletedInfo.nFiles > 0:
            deletedInfofile = self.infopath(deleted=True)
            #debug_log("SimpleFilesetCache(%s)::saveDeletions deletedInfo\n" % self._path)
            try:
                with open(deletedInfofile, 'w') as f:
                    self._deletedInfo.write(f)
            except IOError:
                warning("can't write deleted info %s, ignoring" % deletedInfofile)
                self._deletedInfo = FilesetInfo()



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
from .util import str2size, size2str0, warning
from .Buckets import Buckets
from .FilesetInfo import FilesetInfo

class FilesetInfoAccumulator(object):

    def __init__(self, attrs):
        self._total = FilesetInfo()
        self._users = {}
        self._datasets = {}
        self._sizebuckets = Buckets([str2size(s) for s in attrs['sizebuckets']] if 'sizebuckets' in attrs else [])
        self._sizes = [None] * self._sizebuckets.len

    def add(self, filespec):
        self._total.add(1, filespec.size)

        # user
        if filespec.user in self._users:
            user = self._users[filespec.user]
        else:
            user = FilesetInfo()
            self._users[filespec.user] = user
        user.add(1, filespec.size)

        # dataset
        if filespec.dataset in self._datasets:
            dataset = self._datasets[filespec.dataset]
        else:
            dataset = FilesetInfo()
            self._datasets[filespec.dataset] = dataset
        dataset.add(1, filespec.size)

        # size
        i = self._sizebuckets.indexContaining(filespec.size)
        sizes0 = self._sizes[i]
        if sizes0 is None:
            sizes0 = FilesetInfo()
            self._sizes[i] = sizes0
        sizes0.add(1, filespec.size)

    def accumulate(self, info, sel):
        self._total.add(info.nFiles, info.totalSize)
        if sel.owner is not None:
            if sel.owner not in self._users:
                user0 = FilesetInfo()
                self._users[sel.owner] = user0
            else:
                user0 = self._users[sel.owner]
            user0.add(info.nFiles, info.totalSize)
        if sel.dataset is not None:
            if sel.dataset not in self._datasets:
                dataset0 = FilesetInfo()
                self._datasets[sel.dataset] = dataset0
            else:
                dataset0 = self._datasets[sel.dataset]
            dataset0.add(info.nFiles, info.totalSize)
        if sel.sizebucket is not None:
            i = self._sizebuckets.index(sel.sizebucket)
            sizes0 = self._sizes[i]
            if sizes0 is None:
                sizes0 = FilesetInfo()
                self._sizes[i] = sizes0
            sizes0.add(info.nFiles, info.totalSize)

    def decumulate(self, info, sel):
        self._total.remove(info.nFiles, info.totalSize)
        if sel.owner is not None:
            if sel.owner in self._users:
                user0 = self._users[sel.owner]
                user0.remove(info.nFiles, info.totalSize)
                if user0.nFiles == 0:
                    # remove user, since no files left
                    self._users.pop(sel.owner, None)
        if sel.dataset is not None:
            if sel.dataset in self._datasets:
                dataset0 = self._datasets[sel.dataset]
                dataset0.remove(info.nFiles, info.totalSize)
                if dataset0.nFiles == 0:
                    # remove dataset, since no files left
                    self._datasets.pop(sel.dataset, None)
        if sel.sizebucket is not None:
            i = self._sizebuckets.index(sel.sizebucket)
            sizes0 = self._sizes[i]
            sizes0.remove(info.nFiles, info.totalSize)
            if sizes0.nFiles == 0:
                # remove sizes, since no files left
                self._sizes[i] = None

    def fmt_total(self):
        return "total %s" % str(self._total)

    def fmt_users(self):
        lines = [self.fmt_total()]
        for user in sorted(list(self._users.items()), key=lambda u: u[1].totalSize, reverse=True):
            name = user[0]
            info = user[1]
            # exclude trivial small stuff
            if info.totalSize > 1024:
                lines.append("%s %s" % (name, str(info)))
        return '\n'.join(lines)

    def iterusers(self):
        return iter(self._users.items())

    def fmt_datasets(self):
        lines = [self.fmt_total()]
        for dataset in sorted(list(self._datasets.items()), key=lambda u: u[1].totalSize, reverse=True):
            name = dataset[0]
            info = dataset[1]
            # exclude trivial small stuff
            if info.totalSize > 1024:
                lines.append("%s %s" % (name, str(info)))
        return '\n'.join(lines)

    def fmt_sizes(self):
        lines = [self.fmt_total()]
        last = len(self._sizes) - 1
        for i in range(last + 1):
            if self._sizes[i] is not None:
                info = self._sizes[i]
                # exclude trivial small stuff
                if info.totalSize > 1024:
                    if i < last:
                        interval = "%4s - %4s" % (size2str0(self._sizebuckets.bound(i)), size2str0(self._sizebuckets.bound(i + 1)))
                    else:
                        interval = "%4s +     " % size2str0(self._sizebuckets.bound(i))
                    lines.append("%s  %s" % (interval, str(info)))
        return '\n'.join(lines)



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
import datetime
import os.path
import re
import shutil

from .Filter import Filter
from .FilespecMerger import FilespecMerger
from .PooledFile import listdir
from .util import verbose_stderr, debug_log

class WeeklyFilesetCache(object):

    @classmethod
    def week(cls, t):
        """Return time t as an integer valued week YYYYWW."""
        dt = datetime.datetime.fromtimestamp(t)
        isoyear,isoweek,isoweekday = dt.isocalendar()
        return isoyear * 100 + isoweek

    def __init__(self, path, deltadir, mapper, attrs, sel, next):
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs
        self._sel = sel
        self._next = next
        self._weeks = {}        # of fileset, indexed by integer week

        # load stubs for all weeks found
        if os.path.exists(self._path):
            for wstr in listdir(self._path):
                w = int(wstr)
                self._weeks[w] = None # stub

    def _subpath(self, w):
        return os.path.join(self._path, str(w))

    def _subdeltadir(self, w):
        return os.path.join(self._deltadir, str(w))

    def _fileset(self, w):
        """On demand creation of child filesets."""
        if w in self._weeks:
            fileset = self._weeks[w]
        else:
            fileset = None
        if fileset is None:
            fileset = self._next(self._subpath(w), self._subdeltadir(w), self._mapper, self._attrs, self._sel)
            self._weeks[w] = fileset
        return fileset

    def select(self, filter=None):
        merger = FilespecMerger()
        weeks = sorted(self._weeks.keys())
        for w in weeks:
            if filter is None or filter.mtimeBefore is None or w <= self.__class__.week(filter.mtimeBefore):
                if filter is not None and filter.mtimeBefore is not None and w < self.__class__.week(filter.mtimeBefore):
                    f1 = Filter.clearMtime(filter)
                else:
                    f1 = filter
                merger.add(self._fileset(w).select(f1))
        # no yield from in python 2, so:
        for filespec in merger.merge():
            yield filespec

    def merge_info(self, acc, filter=None):
        #debug_log("WeeklyFilesetCache(%s) merge_info\n" % self._path)
        for w in list(self._weeks.keys()):
            if filter is None or filter.mtimeBefore is None or w <= self.__class__.week(filter.mtimeBefore):
                if filter is not None and filter.mtimeBefore is not None and w < self.__class__.week(filter.mtimeBefore):
                    f1 = Filter.clearMtime(filter)
                else:
                    f1 = filter
                self._fileset(w).merge_info(acc, f1)

    def create(self):
        """Create empty cache on disk, purging any previous."""
        #debug_log("WeeklyFilesetCache creating at %s\n" % self._path)
        if os.path.exists(self._path):
            # Purge existing cache.
            # For safety in case of misconfiguration, we only delete directories in the format YYYYWW
            YYYYWW = re.compile(r"""^\d\d\d\d\d\d$""")
            for x in listdir(self._path):
                px = os.path.join(self._path, x)
                if YYYYWW.match(x):
                    shutil.rmtree(px)
                else:
                    verbose_stderr("WARNING: cache purge ignoring %s\n" % px)
        else:
            os.makedirs(self._path)
        self._weeks = {}

    def add(self, filespec):
        w = self.__class__.week(filespec.mtime)
        fileset = self._fileset(w)
        fileset.add(filespec)

    def finalize(self):
        for w in self._weeks.values():
            if w is not None:
                w.finalize()

    def saveDeletions(self):
        #debug_log("WeeklyFilesetCache(%s)::saveDeletions\n" % self._path)
        for w in self._weeks.values():
            if w is not None:
                w.saveDeletions()



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from .Fileset import Fileset
from .FilespecMerger import FilespecMerger

class UnionFileset(Fileset):

    def __init__(self, name, filesets):
        self.name = name
        self._filesets = filesets

    def description(self):
        return "%s union %s" % (self.name, ' '.join(sorted([fileset.name for fileset in self._filesets])))

    def select(self, filter=None):
        merger = FilespecMerger()
        for fileset in self._filesets:
            merger.add(fileset.select(filter))
        # no yield from in python 2, so:
        for filespec in merger.merge():
            yield filespec

    def merge_info(self, acc, filter=None):
        for fileset in self._filesets:
            fileset.merge_info(acc, filter)



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import grp
import pwd
import re

class Pathway(object):

    def __init__(self):
        self._datasetRegex = None
        self._datasetReplace = None
        self._ignorePathRegexes = []

    def setDatasetRegex(self, datasetRegex, datasetReplace):
        self._datasetRegex = re.compile(datasetRegex)
        self._datasetReplace = datasetReplace

    def clearDatasetRegex(self):
        self._datasetRegex = None
        self._datasetReplace = None

    def datasetFromPath(self, path):
        noDatasetFound = '-'
        if self._datasetRegex is None:
            return noDatasetFound
        dataset, n = re.subn(self._datasetRegex, self._datasetReplace, path, 1)
        if n == 1:
            return dataset
        else:
            return noDatasetFound

    def setIgnorePathsFrom(self, ignorefilepath):
        self._ignorePathRegexes = []
        with open(ignorefilepath) as f:
            for line in f:
                i_hash = line.find('#')
                if i_hash != -1:
                    regex = line[:i_hash].strip()
                else:
                    regex = line.strip()
                if regex != '':
                    self._ignorePathRegexes.append(re.compile(regex))

    def ignored(self, path):
        for r in self._ignorePathRegexes:
            if re.search(r, path):
                return True
        return False



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

import errno
import os
import os.path
import grp
import pwd
import re
import stat

from .CLIError import CLIError
from .Fileset import Fileset
from .Filespec import Filespec
from .PooledFile import listdir
from .util import filemode, verbose_stderr

class FindFileset(Fileset):

    @classmethod
    def parse(cls, mapper, pathway, name, toks):
        if len(toks) == 1:
            path = toks[0]
            match = '^'
            replace = toks[0].rstrip('/') + '/'
        elif len(toks) == 3:
            path = toks[0]
            match = toks[1]
            replace = toks[2]
        else:
            raise CLIError("find requires path, and either both of match-re, replace-str or neither")
        return cls(mapper, pathway, name, path, match, replace)

    def __init__(self, mapper, pathway, name, path, match, replace):
        #print("FindFileset init '%s' '%s' '%s'" % (path, match, replace))
        Fileset.__init__(self)
        self._mapper = mapper
        self._pathway = pathway
        self.name = name
        self._path = path
        self._match = match
        self._replace = replace

    def description(self):
        return "%s directory %s" % (self.name, self._path)

    def select(self, filter=None):
        verbose_stderr("fileset %s scanning files under %s\n" % (self.name, self._path))
        pathlen = len(self._path) + (0 if self._path[-1] == '/' else 1)
        dirs = [self._path]
        # can't use os.walk, as that fails if we hit too many open files
        while dirs != []:
            root = dirs[0]
            relroot = root[pathlen:]
            dirs = dirs[1:]
            try:
                entries = listdir(root)
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # ignore just-disappeared directory
                    continue
                else:
                    raise

            for x in entries:
                xpath = os.path.join(root, x)
                xrel = os.path.join(relroot, x)
                try:
                    s = os.lstat(xpath)
                except OSError as e:
                    if e.errno == errno.ENOENT:
                        # ignore just-disappeared path
                        continue
                    else:
                        raise
                if stat.S_ISDIR(s.st_mode):
                    dirs.append(xpath)
                path = re.sub(self._match, self._replace, xrel)
                filespec = Filespec(fileset=self,
                                    dataset=self._pathway.datasetFromPath(path),
                                    path=path,
                                    user=self._mapper.usernameFromId(s.st_uid),
                                    group=self._mapper.groupnameFromId(s.st_gid),
                                    size=s.st_size,
                                    mtime=s.st_mtime,
                                    perms=filemode(s.st_mode))
                if filter == None or filter.selects(filespec):
                    #print("FindFileset scan found %s" % filespec)
                    yield filespec



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
from .FilesetInfoAccumulator import FilesetInfoAccumulator
from .FilesetSelector import FilesetSelector

class Fileset(object):
    """Fileset is a base class."""

    def __init__(self):
        pass

    def info(self, attrs, filter=None):
        acc = FilesetInfoAccumulator(attrs)
        self.merge_info(acc, filter)
        return acc

    def sorted(self, filter=None, sorter=None):
        if sorter is None:
            for filespec in self.select(filter):
                yield filespec
        else:
            filespecs = []
            for filespec in self.select(filter):
                filespecs.append(filespec)
            filespecs.sort(key=sorter.key())
            for filespec in filespecs:
                yield filespec

    def delete(self, filespec):
        pass



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

import os
import os.path
import re
import readline
import shlex
import time

from .CLIError import CLIError
from .Filter import Filter
from .Sorter import Sorter
from .Grouper import Grouper
from .util import str2size

def parseCommandOptions(now, toks, filter=False, sorter=False, grouper=False):
    """Parse filter and/or sorter options."""

    # filter options
    owner = None
    dataset = None
    sizeGeq = None
    mtimeBefore = None
    notPaths = []

    # sorter options
    bySize = False

    # grouper options
    collapse = None

    i = 0
    while i < len(toks):
        tok = toks[i]
        if tok == '-user' and filter:
            if i + 1 >= len(toks):
                raise CLIError("-user missing parameter")
            if owner is not None:
                raise CLIError("duplicate -user")
            owner = toks[i + 1]
            i += 1
        elif tok == '-dataset' and filter:
            if i + 1 >= len(toks):
                raise CLIError("-dataset missing parameter")
            if dataset is not None:
                raise CLIError("duplicate -dataset")
            dataset = toks[i + 1]
            i += 1
        elif tok == '-size' and filter:
            if i + 1 >= len(toks):
                raise CLIError("-size missing parameter")
            size = toks[i + 1]
            if len(size) < 2 or size[0] != '+':
                raise CLIError("-size only supports +n[kMGT] format")
            n = str2size(size[1:])
            if sizeGeq is not None:
                raise CLIError("duplicate -size")
            sizeGeq = n
            i += 1
        elif tok == '-mtime' and filter:
            if i + 1 >= len(toks):
                raise CLIError("-mtime missing parameter")
            mtime = toks[i + 1]
            if len(mtime) < 2 or mtime[0] != '+':
                raise CLIError("-mtime only supports +n format")
            try:
                n = int(mtime[1:])
            except ValueError:
                raise CLIError("-mtime only supports +n format")
            if mtimeBefore is not None:
                raise CLIError("duplicate -mtime")
            mtimeBefore = now - n * 60 * 60 * 24
            i += 1
        elif tok == '!' and filter:
            # only for -path
            if i + 2 >= len(toks) or toks[i + 1] != '-path':
                raise CLIError("! requires -path <glob>")
            notPaths.append(toks[i + 2])
            i += 2
        elif tok == '-by-size' and sorter:
            bySize = True
        elif tok == '-depth' and grouper:
            if i + 1 >= len(toks):
                raise CLIError("-depth missing parameter")
            try:
                collapse = int(toks[i + 1])
            except ValueError:
                raise CLIError("-depth requires integer parameter")
            i += 1
        else:
            category = ""
            if filter and not sorter and not grouper:
                category = "filter "
            if sorter and not filter and not grouper:
                category = "sorter "
            if grouper and not sorter and not filter:
                category = "grouper "
            raise CLIError("unknown %sparameter %s" % (category, tok))
        i += 1

    if filter and (owner is not None or dataset is not None or sizeGeq is not None or mtimeBefore is not None or notPaths != []):
        f0 = Filter(owner=owner, dataset=dataset, sizeGeq=sizeGeq, mtimeBefore=mtimeBefore, notPaths=notPaths)
    else:
        f0 = None
    if sorter and bySize:
        s0 = Sorter(bySize=bySize)
    else:
        s0 = None
    if grouper:
        g0 = Grouper(collapse=collapse)
    else:
        g0 = None

    return f0, s0, g0



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from .Fileset import Fileset
from .Filter import Filter
from .util import debug_log

class FilterFileset(Fileset):

    def __init__(self, name, fileset, filter):
        #debug_log("FilterFileset(%s), filter=%s\n" % (name, filter))
        self.name = name
        self._fileset = fileset
        self._filter = filter

    def description(self):
        return "%s filter %s %s" % (self.name, self._fileset.name, self._filter)

    def select(self, filter=None):
        f1 = self._filter.intersect(filter)
        #debug_log("FilterFileset(%s)::select filter=%s\n" % (self.name, f1))
        for filespec in self._fileset.select(f1):
            yield filespec

    def merge_info(self, acc, filter=None):
        f1 = self._filter.intersect(filter)
        #debug_log("FilterFileset(%s)::select filter=%s\n" % (self.name, f1))
        self._fileset.merge_info(acc, f1)



from __future__ import print_function
from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
import calendar
import errno
import os
import time

from .util import fbTimeFmt, time2str, date2str, size2str, debug_log

# FileSpec fields:
# path - string, relative to some (externally defined) rootdir
# user - string
# group - string
# size - in bytes
# mtime - seconds since epoch
# perms - string, in ls -l format
class Filespec(object):

    @classmethod
    def fromFile(cls, f, fileset, sel):
        for line in f:
            fields = line.rstrip().split(None, 4)
            if len(fields) != 5:
                print("bad filespec: %s" % line.rstrip())
            else:
                yield Filespec(fileset,
                               sel.dataset,
                               fields[4],
                               sel.owner,
                               fields[0],
                               int(fields[1]),
                               calendar.timegm(time.strptime(fields[2], fbTimeFmt)),
                               fields[3])

    def __init__(self, fileset, dataset, path, user, group, size, mtime, perms):
        self.fileset = fileset  # fileset whicih owns this filespec
        self.dataset = dataset
        self.path = path
        self.user = user
        self.group = group
        self.size = size
        self.mtime = mtime
        self.perms = perms

    def __str__(self):
        return self.format(all=True)[0]

    def isdir(self):
        return self.perms[0] == 'd'

    def delete(self, logf):
        """Delete file or directory, logging path to logf on success."""
        deleted = False
        try:
            if self.isdir():
                os.rmdir(self.path)
            else:
                os.remove(self.path)
            deleted = True
            # tell owning fileset we're deleted
            self.fileset.delete(self)
        except OSError as e:
            debug_log("failed to delete %s, %s\n" % (self.path, e))
            if e.errno == errno.ENOENT:
                # deleted already, don't care
                pass
            elif e.errno == errno.EACCES or e.errno == errno.EPERM:
                # silently refuse to delete where we don't have permission
                pass
            elif e.errno == errno.ENOTEMPTY:
                # silently refuse to delete non-empty directory
                pass
            else:
                raise
        if deleted:
            logf.write("%s%s\n" % self.format(pad=False))

    @classmethod
    def formattedToPath(cls, formatted):
        """Return the file path from a formatted Filespec (with all=False)."""
        return formatted.split(None, 4)[4]

    def format(self, width=0, all=False, pad=True):
        usergroup = "%s:%s" % (self.user, self.group)
        if len(usergroup) >= width:
            width = len(usergroup) + 1
        if not pad:
            width = 0
        s = ("%s %-" + str(width) + "s %s %4s %s") % (
            self.perms,
            usergroup,
            date2str(self.mtime),
            size2str(self.size),
            self.path)
        if all:
            s += " %s" % (self.dataset)
        return s, width

    def write(self, f):
        f.write("%s %d %s %s %s\n" % (
            self.group,
            self.size,
            time2str(self.mtime),
            self.perms,
            self.path))



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import errno
import os
import time

from .CLIError import CLIError
from .util import date2str, debug_log

class DeletionLog(object):
    """A logfile for deleted files/directories."""

    def __init__(self, attrs):
        # first try the syslogdir, if we have permission
        if 'syslogdir' not in attrs:
            raise CLIError("missing attr syslogdir")
        syslogdirs = attrs['syslogdir']
        if len(syslogdirs) != 1:
            raise CLIError("botched attr syslogdir")
        syslogdir = syslogdirs[0]
        if 'userlogdir' not in attrs:
            raise CLIError("missing attr userlogdir")
        userlogdirs = attrs['userlogdir']
        if len(userlogdirs) != 1:
            raise CLIError("botched attr userlogdir")
        userlogdir = userlogdirs[0]
        try:
            os.makedirs(syslogdir)
        except:
            # ignore errors, we'll catch it in a moment
            pass
        datestamp = date2str(time.time())
        try:
            self._file = open(os.path.join(syslogdir, datestamp), 'a')
        except IOError as e:
            if e.errno == errno.EACCES or e.errno == errno.EPERM or e.errno == errno.ENOENT:
                # syslogdir no good, try userlogdir
                try:
                    os.makedirs(userlogdir)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise
                self._file = open(os.path.join(userlogdir, datestamp), 'a')
            else:
                raise

    def __enter__(self):
        return self._file

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.close()



from __future__ import absolute_import
from __future__ import division
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
from past.utils import old_div
import math
import time

from .util import progress_stderr

class PercentageProgress(object):

    def __init__(self, prefix):
        self._fmt = "\r" + prefix + " [%.1f%% complete%s]"
        self._start = time.time()
        self._progress = -1
        self._maxlen = 0
        self.report(0)

    def report(self, p):
        """Report progress p, from 0 to 1."""
        if p - self._progress > 0.001:
            self._progress = p
            if p > 0:
                elapsed = time.time() - self._start
                t = max(math.floor(elapsed * 1.0 / p - elapsed), 0)
                m = old_div(t, 60)
                s = t % 60
                remaining = ", %d:%02d remaining" % (m, s)
            else:
                remaining = ""
            message = self._fmt % (p * 100, remaining)
            progress_stderr(message)
            if len(message) > self._maxlen:
                self._maxlen = len(message)

    def complete(self):
        progress_stderr("\r%s\r" % (' ' * self._maxlen))



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
from .util import size2str, warning

class FilesetInfo(object):

    @classmethod
    def fromFile(cls, f):
        fi = None
        for line in f:
            try:
                fields = line.rstrip().split(None)
                if fi is None and len(fields) == 2:
                    nFiles = int(fields[0])
                    totalSize = int(fields[1])
                    fi = cls(nFiles, totalSize)
                else:
                    raise ValueError
            except ValueError:
                warning("ignoring bad info line: %s" % line.rstrip())
        return fi

    def __init__(self, nFiles=0, totalSize=0):
        self.nFiles = nFiles
        self.totalSize = totalSize

    def add(self, nFiles, totalSize):
        self.nFiles += nFiles
        self.totalSize += totalSize

    def remove(self, nFiles, totalSize):
        self.nFiles -= nFiles
        self.totalSize -= totalSize

    def __str__(self):
        return "%6s in %9d files" % (size2str(self.totalSize), self.nFiles)

    def write(self, f):
        f.write("%d %d\n" % (self.nFiles, self.totalSize))



from __future__ import absolute_import
# Copyright 2018 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import next
from builtins import range
from builtins import object
import calendar
import os
import time

from .util import fbTimeFmt, time2str, date2str, size2str

class FilespecMerger(object):
    """Merge filespecs from multiple iterators in order of path."""

    def __init__(self):
        self._iters = []

    def add(self, iter):
        """Add an iterator, to be merged in with the others."""
        self._iters.append(iter)

    def merge(self):
        """Yield from all the iterators in order."""
        # get initial value for each iterator
        values = []
        for iter in self._iters:
            try:
                value = next(iter)
            except StopIteration:
                value = None
            values.append(value)

        # merge least value each time until done
        done = False
        while not done:
            # find min value
            min_i = None
            for i in range(len(values)):
                if values[i] is not None and (min_i is None or values[i].path < values[min_i].path):
                    min_i = i

            if min_i is not None:
                yield values[min_i]
                # get next value for iterator which returned this one
                try:
                    values[min_i] = next(self._iters[min_i])
                except StopIteration:
                    values[min_i] = None
            else:
                done = True



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
import errno
import functools
import os.path

from .CLIError import CLIError
from .DatasetFilesetCache import DatasetFilesetCache
from .Fileset import Fileset
from .FilesetSelector import FilesetSelector
from .PooledFile import PooledFile
from .SimpleFilesetCache import SimpleFilesetCache
from .SizeFilesetCache import SizeFilesetCache
from .UserFilesetCache import UserFilesetCache
from .WeeklyFilesetCache import WeeklyFilesetCache
from .util import filedatestr, filetimestr, verbose_stderr, debug_log, progress_stderr, warning

# Stack up the caches we support, so that each cache can instantiate
# its next one, via its next parameter.
class Cache(Fileset):

    def __init__(self, name, fileset, path, deltadir, mapper, attrs):
        Fileset.__init__(self)
        self.name = name
        self._fileset = fileset
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs.copy() # copy the dictionary, so we freeze its values
        self._cache0 = None
        self._caches = [WeeklyFilesetCache, SizeFilesetCache, DatasetFilesetCache, UserFilesetCache]

    def _exists(self):
        try:
            s = os.stat(self._path)
            exists = True
        except OSError as e:
            if e.errno == errno.ENOENT:
                exists = False
            else:
                raise
        return exists

    def _abortIfMissingCache(self):
        if not self._exists():
            raise CLIError("<missing-cache>, use 'update-cache %s', or 'update-cache' for all" % self.name)

    def description(self):
        if self._exists():
            return "%s cached on %s" % (self._fileset.description(), filedatestr(self._path))
        else:
            return "%s <missing-cache>" % (self._fileset.description())

    def _cache(self):
        if self._cache0 is None:
            self._cache0 = self._newcache(self._path, self._deltadir, self._mapper, self._attrs, FilesetSelector(), 0)
        return self._cache0

    def _newcache(self, path, deltadir, mapper, attrs, sel, level):
        if level < len(self._caches):
            return self._caches[level](path, deltadir, mapper, attrs, sel, functools.partial(self._newcache, level = level + 1))
        else:
            return SimpleFilesetCache(path, deltadir, mapper, attrs, sel)

    def select(self, filter=None):
        self._abortIfMissingCache()
        cache = self._cache()
        filterStr = " " + str(filter) if filter is not None else ""
        verbose_stderr("fileset %s%s reading from %s cache at %s\n" % (self.name, filterStr, filetimestr(self._path), self._path))
        for filespec in cache.select(filter):
            yield filespec

    def merge_info(self, acc, filter=None):
        self._abortIfMissingCache()
        #debug_log("Cache(%s) merge_info\n" % self.name)
        cache = self._cache()
        cache.merge_info(acc, filter)

    def update(self):
        cache = self._cache()
        try:
            cache.create()
        except OSError as e:
            if e.errno == errno.EACCES or e.errno == errno.EPERM:
                # not ours to update, so silently do nothing
                warning("can't update system cache %s" % self.name)
                return
        for filespec in self._fileset.select():
            cache.add(filespec)
        # ensure from here on we don't hit open file problems
        PooledFile.flushAll()
        cache.finalize()
        # touch cache rootdir, to show updated
        os.utime(self._path, None)
        progress_stderr("updated %s\n" % self.name)

    def saveDeletions(self):
        #debug_log("Cache(%s)::saveDeletions\n" % self.name)
        cache = self._cache()
        cache.saveDeletions()



#!/usr/bin/env python
#
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import os.path
import readline
import sys

from filebutler.CLI import CLI
from filebutler.version import get_version

def main():
    parser = argparse.ArgumentParser(description="""filebutler v%s - view and prune old files.""" % get_version())
    parser.add_argument('--batch', action='store_true', dest='batch', help='batch mode')
    parser.add_argument('--debug', type=argparse.FileType('w'), dest='debug', metavar='FILE', help='debug mode')
    parser.add_argument('--profile', action='store_true', dest='profile', help='profile mode')
    parser.add_argument('--version', action='store_true', dest='version', help='show version and exit')
    parser.add_argument('-v', action='store_true', dest='verbose', help='verbose mode')
    parser.add_argument('-c', dest='command', help='command to invoke')
    args = parser.parse_args()

    if args.version:
        print('filebutler v%s' % get_version())
        sys.exit(0)

    cli = CLI(args)
    cli.startup()

    if args.command is not None:
        cli.execute(args.command.split(';'))
    else:
        historyFile = os.path.expanduser('~/.filebutler.history')
        if os.path.exists(historyFile):
            readline.read_history_file(historyFile)
        cli.interact()
        readline.write_history_file(historyFile)

if __name__ == '__main__':
    main()



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
import datetime
import pytz
import tzlocal

class Localtime(object):

    def __init__(self):
        self._epoch = datetime.datetime.utcfromtimestamp(0)
        self._localtimezone = pytz.timezone(str(tzlocal.get_localzone()))

    def datetime(self, *args):
        return datetime.datetime(*args)

    def t(self, *args):
        dt = self.datetime(*args)
        return (dt - self._epoch).total_seconds()



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
import grp
import pwd
import re

class Mapper(object):

    def __init__(self):
        self._uids = {}
        self._usernames = {}
        self._groupnames = {}
        self._datasetRegex = None
        self._datasetReplace = None

    def uidFromUsername(self, username):
        """Returns -1 if can't resolve the username."""
        if username not in self._uids:
            try:
                pw = pwd.getpwnam(username)
                uid = pw[2]
            except KeyError:
                try:
                    uid = int(username)
                except ValueError:
                    uid = -1
            self._uids[username] = uid
        else:
            uid = self._uids[username]
        return uid

    def usernameFromId(self, uid):
        if uid not in self._usernames:
            try:
                pw = pwd.getpwuid(uid)
                name = pw[0]
            except KeyError:
                name = str(uid)
            self._usernames[uid] = name
        else:
            name = self._usernames[uid]
        return name

    def usernameFromString(self, s):
        try:
            uid = int(s)
            return self.usernameFromId(uid)
        except ValueError:
            return s

    def groupnameFromId(self, gid):
        if gid not in self._groupnames:
            try:
                gr = grp.getgrgid(gid)
                name = gr[0]
            except KeyError:
                name = str(gid)
            self._groupnames[gid] = name
        else:
            name = self._groupnames[gid]
        return name

    def groupnameFromString(self, s):
        try:
            gid = int(s)
            return self.groupnameFromId(gid)
        except ValueError:
            return s

    def setDatasetRegex(self, datasetRegex, datasetReplace):
        self._datasetRegex = datasetRegex
        self._datasetReplace = datasetReplace

    def clearDatasetRegex(self):
        self._datasetRegex = None
        self._datasetReplace = None

    def datasetFromPath(self, path):
        noDatasetFound = '-'
        if self._datasetRegex is None:
            return noDatasetFound
        dataset, n = re.subn(self._datasetRegex, self._datasetReplace, path, 1)
        if n == 1:
            return dataset
        else:
            return noDatasetFound



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import os
import os.path

from .Filter import Filter
from .FilespecMerger import FilespecMerger
from .PooledFile import listdir
from .util import debug_log

class UserFilesetCache(object):

    def __init__(self, path, deltadir, mapper, attrs, sel, next):
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs
        self._sel = sel
        self._next = next
        self._users = {}        # of fileset, indexed by integer user
        self._permissioned = {}        # of boolean, indexed by integer user

        # load stubs for all users found
        if os.path.exists(self._path):
            for u in listdir(self._path):
                self._users[u] = None # stub
                self._permissioned[u] = False

    def _subpath(self, u):
        return os.path.join(self._path, u)

    def _subdeltadir(self, u):
        return os.path.join(self._deltadir, u)

    def _fileset(self, u):
        """On demand creation of child filesets."""
        if u in self._users:
            fileset = self._users[u]
        else:
            fileset = None
        if fileset is None:
            fileset = self._next(self._subpath(u), self._subdeltadir(u), self._mapper, self._attrs, self._sel.withOwner(u))
            self._users[u] = fileset
            self._permissioned[u] = False
        return fileset

    def select(self, filter=None):
        merger = FilespecMerger()
        users = sorted(self._users.keys())
        for u in users:
            if filter is None or filter.owner is None or u == filter.owner:
                merger.add(self._fileset(u).select(Filter.clearOwner(filter)))
        # no yield from in python 2, so:
        for filespec in merger.merge():
            yield filespec

    def merge_info(self, acc, filter=None):
        #debug_log("UserFilesetCache(%s) merge_info\n" % self._path)
        for u in list(self._users.keys()):
            if filter is None or filter.owner is None or u == filter.owner:
                self._fileset(u).merge_info(acc, Filter.clearOwner(filter))

    def add(self, filespec):
        fileset = self._fileset(filespec.user)
        fileset.add(filespec)
        if 'private' in self._attrs and os.geteuid() == 0 and not self._permissioned[filespec.user]:
            # set permissions of fileset directory
            upath = self._subpath(filespec.user)
            uid = self._mapper.uidFromUsername(filespec.user)
            if uid != -1:
                os.chown(upath, uid, -1)
            os.chmod(upath, 0o500)
            self._permissioned[filespec.user] = True

    def finalize(self):
        for u in self._users.values():
            if u is not None:
                u.finalize()

    def saveDeletions(self):
        #debug_log("UserFilesetCache(%s)::saveDeletions\n" % self._path)
        for u in self._users.values():
            if u is not None:
                u.saveDeletions()



from __future__ import absolute_import
from __future__ import division
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from past.utils import old_div
import errno
import os
import resource
import time
from stat import *
import sys

from .CLIError import CLIError

_debug = None
_profile = False
_progress = True
_verbose = False
def initialize(args):
    global _debug
    global _profile
    global _verbose
    global _progress
    if args.verbose:
        _verbose = True
    if args.profile:
        _profile = True
        _verbose = True
    if args.batch:
        _progress = False
    if args.debug:
        _debug = args.debug
        _verbose = True
        verbose_stderr("debug mode\n")

def stderr(msg):
    sys.stderr.write(msg)
def debug_log(msg):
    if _debug is not None:
        _debug.write(msg)
def progress_stderr(msg):
    if _progress:
        stderr(msg)
def verbose_stderr(msg):
    if _verbose or _debug:
        stderr(msg)
def warning(msg):
    stderr("warning: %s\n" % msg)

def profile():
    return _profile

fbTimeFmt = "%Y%m%d-%H%M%S"
fbDateFmt = "%Y-%m-%d"

def time2str(t):
    return time.strftime(fbTimeFmt, time.localtime(t))

def date2str(t):
    return time.strftime(fbDateFmt, time.localtime(t))

# stolen from Python 3:
_filemode_table = (
    ((S_IFLNK,         "l"),
     (S_IFREG,         "-"),
     (S_IFBLK,         "b"),
     (S_IFDIR,         "d"),
     (S_IFCHR,         "c"),
     (S_IFIFO,         "p")),

    ((S_IRUSR,         "r"),),
    ((S_IWUSR,         "w"),),
    ((S_IXUSR|S_ISUID, "s"),
     (S_ISUID,         "S"),
     (S_IXUSR,         "x")),

    ((S_IRGRP,         "r"),),
    ((S_IWGRP,         "w"),),
    ((S_IXGRP|S_ISGID, "s"),
     (S_ISGID,         "S"),
     (S_IXGRP,         "x")),

    ((S_IROTH,         "r"),),
    ((S_IWOTH,         "w"),),
    ((S_IXOTH|S_ISVTX, "t"),
     (S_ISVTX,         "T"),
     (S_IXOTH,         "x"))
)

def filemode(mode):
    """Convert a file's mode to a string of the form '-rwxrwxrwx'."""
    perm = []
    for table in _filemode_table:
        for bit, char in table:
            if mode & bit == bit:
                perm.append(char)
                break
        else:
            perm.append("-")
    return "".join(perm)

Kilo = 1024
Mega = 1024 ** 2
Giga = 1024 ** 3
Tera = 1024 ** 4

def size2str0(n):
    if n == 0:
        return "0"
    else:
        return size2str(n)

def size2str(n):
    if n < Kilo:
        return "0k"             # so small we don't care
    elif n < Mega:
        return "%dk" % (old_div(n, Kilo))
    elif n < Giga:
        return "%dM" % (old_div(n, Mega))
    elif n < Tera:
        return "%dG" % (old_div(n, Giga))
    else:
        return "%.1fT" % (n * 1.0 / Tera)

def str2size(s):
    unit = s[-1]
    x = int(s[:-1])
    if unit == 'k':
        return x * Kilo
    elif unit == 'M':
        return x * Mega
    elif unit == 'G':
        return x * Giga
    elif unit == 'T':
        return x * Tera
    else:
        raise CLIError("unsupported filesize %s, must be n[kMGT]" % s)

def filetimestr(path):
    try:
        return time2str(os.stat(path).st_mtime)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return "<missing>"
        else:
            raise

def filedatestr(path):
    try:
        return date2str(os.stat(path).st_mtime)
    except OSError as e:
        if e.errno == errno.ENOENT:
            return "<missing>"
        else:
            raise

# adapted from https://gist.github.com/turicas/5278558
def unix_time(function, args=tuple(), kwargs={}):
    '''Return `real`, `sys` and `user` elapsed time, like UNIX's command `time`
    You can calculate the amount of used CPU-time used by your
    function/callable by summing `user` and `sys`. `real` is just like the wall
    clock.
    Note that `sys` and `user`'s resolutions are limited by the resolution of
    the operating system's software clock (check `man 7 time` for more
    details).
    '''
    start_time, start_resources = time.time(), resource.getrusage(resource.RUSAGE_SELF)
    function(*args, **kwargs)
    end_resources, end_time = resource.getrusage(resource.RUSAGE_SELF), time.time()

    return {'real': end_time - start_time,
            'sys': end_resources.ru_stime - start_resources.ru_stime,
            'user': end_resources.ru_utime - start_resources.ru_utime}



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

import pkg_resources

def get_version():
    package = __name__.split('.', 1)[0]
    try:
        version = pkg_resources.get_distribution(package).version
    except pkg_resources.DistributionNotFound:
        # package is not installed
        version = 'UNDEFINED'
    return version



from __future__ import absolute_import
from __future__ import division
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import object
from past.utils import old_div
import datetime
import fnmatch

from .util import Giga, date2str, debug_log

def liberal(fn, a, b):
    if a is None:
        return b
    elif b is None:
        return a
    else:
        return fn(a, b)

class Filter(object):

    @classmethod
    def clearOwner(cls, f0):
        """Return a copy of f0 with no owner specified, or None if f0 is None."""
        if f0 is None or f0.dataset is None and f0.sizeGeq is None and f0.mtimeBefore is None and f0.notPaths == []:
            return None
        else:
            return cls(None, f0.dataset, f0.sizeGeq, f0.mtimeBefore, f0.notPaths)

    @classmethod
    def clearDataset(cls, f0):
        """Return a copy of f0 with no dataset specified, or None if f0 is None."""
        if f0 is None or f0.owner is None and f0.sizeGeq is None and f0.mtimeBefore is None and f0.notPaths == []:
            return None
        else:
            return cls(f0.owner, None, f0.sizeGeq, f0.mtimeBefore, f0.notPaths)

    @classmethod
    def clearMtime(cls, f0):
        """Return a copy of f0 with no mtime specified, or None if f0 is None."""
        if f0 is None or f0.owner is None and f0.dataset is None and f0.sizeGeq is None and f0.notPaths == []:
            return None
        else:
            return cls(f0.owner, f0.dataset, f0.sizeGeq, None, f0.notPaths)

    @classmethod
    def clearSize(cls, f0):
        """Return a copy of f0 with no size specified, or None if f0 is None."""
        if f0 is None or f0.owner is None and f0.dataset is None and f0.mtimeBefore is None and f0.notPaths == []:
            return None
        else:
            return cls(f0.owner, f0.dataset, None, f0.mtimeBefore, f0.notPaths)

    def __init__(self, owner=None, dataset=None, sizeGeq=None, mtimeBefore=None, notPaths=[]):
        self.owner = owner
        self.dataset = dataset
        self.sizeGeq = sizeGeq
        self.mtimeBefore = mtimeBefore
        self.notPaths = notPaths

    def __str__(self):
        s = ""
        def append(s0, s1):
            if len(s0) > 0:
                return s0 + ',' + s1
            else:
                return s1
        if self.owner is not None:
            s = append(s, "owner:%s" % self.owner)
        if self.dataset is not None:
            s = append(s, "dataset:%s" % self.dataset)
        if self.sizeGeq is not None:
            s = append(s, "size:+%dG" % (old_div(self.sizeGeq, Giga)))
        if self.mtimeBefore is not None:
            s = append(s, "older:%s" % date2str(self.mtimeBefore))
        if len(self.notPaths) > 0:
            s = append(s, "!paths:%s" % str(self.notPaths))
        return s

    def intersect(self, f1):
        """Return a new filter which is the intersection of self with the parameter f1."""
        if f1 is None:
            return self

        if self.owner is None:
            owner = f1.owner
        elif f1.owner is not None:
            if self.owner == f1.owner:
                owner = self.owner
            else:
                # incompatible, so set to something impossible, which we test for in selects()
                owner = "%s+%s" % (self.owner, f1.owner)
        else:
            owner = self.owner
        if self.dataset is None:
            dataset = f1.dataset
        elif f1.dataset is not None:
            if self.dataset == f1.dataset:
                dataset = self.dataset
            else:
                # incompatible, so set to something impossible, which we test for in selects()
                dataset = '/'
        else:
            dataset = self.dataset
        sizeGeq = liberal(max, self.sizeGeq, f1.sizeGeq)
        mtimeBefore = liberal(min, self.mtimeBefore, f1.mtimeBefore)
        notPaths = self.notPaths + f1.notPaths
        f2 = self.__class__(owner, dataset, sizeGeq, mtimeBefore, notPaths)
        #debug_log("Filter(%s).intersect(%s)=%s\n" % (self, f1, f2))
        return f2

    def selects(self, filespec):
        if self.owner is not None and '+' in self.owner:
            return False
        if self.owner is not None and filespec.user != self.owner:
            return False
        if self.dataset == '/':
            return False
        if self.dataset is not None and filespec.dataset != self.dataset:
            return False
        if self.sizeGeq is not None and filespec.size < self.sizeGeq:
            return False
        if self.mtimeBefore is not None and filespec.mtime >= self.mtimeBefore:
            return False
        if len(self.notPaths) > 0:
            for notPath in self.notPaths:
                if fnmatch.fnmatchcase(filespec.path, notPath):
                    return False
        #debug_log("%s selects %s\n" % (self, filespec.path))
        return True



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import range
from builtins import object
import calendar
import datetime
import os
import re

from .Fileset import Fileset
from .Filespec import Filespec
from .Localtime import Localtime
from .PercentageProgress import PercentageProgress
from .CLIError import CLIError
from .util import warning, verbose_stderr

class GnuFindOutFileset(Fileset):

    @classmethod
    def parse(cls, mapper, pathway, name, toks):
        if len(toks) != 3:
            raise CLIError("find.gnu.out requires path, match-re, replace-str")
        path = toks[0]
        match = toks[1]
        replace = toks[2]
        return cls(mapper, pathway, name, path, match, replace)

    def __init__(self, mapper, pathway, name, path, match, replace):
        #print("GnuFindOutFileset init %s %s %s" % (path, match, replace))
        Fileset.__init__(self)
        self._mapper = mapper
        self._pathway = pathway
        self.name = name
        self._path = path
        self._match = match
        self._replace = replace
        self._dateParser = self.__class__._dateParser()

    def description(self):
        return "%s filelist %s" % (self.name, self._path)

    class _dateParser(object):
        """Converts find format dates into time since epoch."""
        def __init__(self):
            self._localtime = Localtime()
            self._monthNumbers = {}  # indexed by month_abbr, of 1 to 12
            for i in range(12):
                #print calendar.month_abbr[i + 1], i + 1
                self._monthNumbers[calendar.month_abbr[i + 1]] = i + 1
            self._today = datetime.datetime.today()
            #print "today", self._today.year, self._today.month, self._today.day

        def t(self, fields): # fields are as returned by find -ls, so date is in 7, 8, 9.
            month = self._monthNumbers[fields[7]]
            #print "month", month
            if len(fields[9]) == 4:
                #print "third field is year"
                year = int(fields[9])
            else:
                #print "third field is time-of-day"
                if month <= self._today.month:
                    #print "this year"
                    year = self._today.year
                else:
                    #print "last year"
                    year = self._today.year - 1
            #print "year", year
            return self._localtime.t(year, month, int(fields[8]))

    def select(self, filter=None):
        verbose_stderr("fileset %s reading from filelist %s\n" % (self.name, self._path))
        try:
            filesize = os.stat(self._path).st_size
        except OSError as e:
            warning("fileset %s ignoring unreadable filelist %s: %s" % (self.name, self._path, e.strerror))
            return
        progress = PercentageProgress("reading %s" % self._path)
        n_read = 0
        with open(self._path) as f:
            for line in f:
                n_read += len(line)
                progress.report(n_read * 1.0 / filesize)
                fields = line.rstrip().split(None, 10)
                if len(fields) > 10:
                    path = re.sub(self._match, self._replace, fields[10])
                    # strip out symlink destination
                    symlink = path.find(' -> ')
                    if symlink >= 0:
                        path = path[:symlink]
                    filespec = Filespec(fileset=self,
                                        dataset=self._pathway.datasetFromPath(path),
                                        path=path,
                                        user=self._mapper.usernameFromString(fields[4]),
                                        group=self._mapper.groupnameFromString(fields[5]),
                                        size=int(fields[6]),
                                        mtime=self._dateParser.t(fields),
                                        perms=fields[2])
                    if (filter == None or filter.selects(filespec)) and not self._pathway.ignored(filespec.path):
                        #print("GnuFindOutFileset read from file %s" % filespec)
                        yield filespec
        progress.complete()



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
class Sorter(object):

    def __init__(self, bySize=False):
        self.bySize = bySize

    def key(self):
        if self.bySize:
            # negate size, so we sort largest first
            return lambda fs: -fs.size
        else:
            return lambda fs: 0



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

def read_etc_aliases():
    """Read /etc/aliases, and return as a dict."""
    aliases = {}
    with open('/etc/aliases') as f:
        for line in f:
            i_hash = line.find('#')
            if i_hash != -1:
                entry = line[:i_hash]
            else:
                entry = line
            if entry != '':
                toks = entry.split(':', 1)
                if len(toks) == 2:
                    user = toks[0].strip()
                    email = toks[1].strip()
                    if user != '' and email != '':
                        aliases[user] = email
    return aliases



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import os.path

from .Filter import Filter
from .FilespecMerger import FilespecMerger
from .PooledFile import listdir
from .util import debug_log

class DatasetFilesetCache(object):

    def __init__(self, path, deltadir, mapper, attrs, sel, next):
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs
        self._sel = sel
        self._next = next
        self._datasets = {}        # of fileset, indexed by dataset

        # load stubs for all datasets found
        if os.path.exists(self._path):
            for d in listdir(self._path):
                self._datasets[d] = None # stub

    def _subpath(self, d):
        return os.path.join(self._path, d)

    def _subdeltadir(self, d):
        return os.path.join(self._deltadir, d)

    def _fileset(self, d):
        """On demand creation of child filesets."""
        if d in self._datasets:
            fileset = self._datasets[d]
        else:
            fileset = None
        if fileset is None:
            fileset = self._next(self._subpath(d), self._subdeltadir(d), self._mapper, self._attrs, self._sel.withDataset(d))
            self._datasets[d] = fileset
        return fileset

    def select(self, filter=None):
        merger = FilespecMerger()
        datasets = sorted(self._datasets.keys())
        for d in datasets:
            if filter is None or filter.dataset is None or d == filter.dataset:
                merger.add(self._fileset(d).select(Filter.clearDataset(filter)))
        # no yield from in python 2, so:
        for filespec in merger.merge():
            yield filespec

    def merge_info(self, acc, filter=None):
        #debug_log("DatasetFilesetCache(%s) merge_info\n" % self._path)
        for d in list(self._datasets.keys()):
            if filter is None or filter.dataset is None or d == filter.dataset:
                self._fileset(d).merge_info(acc, Filter.clearDataset(filter))

    def add(self, filespec):
        fileset = self._fileset(filespec.dataset)
        fileset.add(filespec)

    def finalize(self):
        for d in self._datasets.values():
            if d is not None:
                d.finalize()

    def saveDeletions(self):
        #debug_log("DatasetFilesetCache(%s)::saveDeletions\n" % self._path)
        for d in self._datasets.values():
            if d is not None:
                d.saveDeletions()



from __future__ import print_function
from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import input
from builtins import str
from builtins import range
from builtins import object
import cProfile
import email.mime.text
import errno
import os
import os.path
import pstats
import re
import readline
import shlex
import smtplib
import string
import time

from .CLIError import CLIError
from .Cache import Cache
from .DeletionLog import DeletionLog
from .Filter import Filter
from .FilterFileset import FilterFileset
from .FindFileset import FindFileset
from .GnuFindOutFileset import GnuFindOutFileset
from .Grouper import Grouper
from .Mapper import Mapper
from .Pager import Pager
from .Pathway import Pathway
from .UnionFileset import UnionFileset
from .aliases import read_etc_aliases
from .options import parseCommandOptions
from .util import stderr, verbose_stderr, debug_log, initialize, profile, unix_time

class CLI(object):

    def __init__(self, args):
        self._attrs = {}
        self._filesetNames = [] # in order of creation
        self._filesets = {}
        self._caches = {}
        self._now = time.time() # for consistency between all filters
        self._mapper = Mapper()
        self._pathway = Pathway()
        self._aliases = read_etc_aliases()
        self.commands = {
            'help':          { 'desc': 'provide help',
                               'usage': 'help',
                               'method': self._helpCmd,
            },
            'quit':          { 'desc': 'finish filebutler session',
                               'usage': 'quit (or Ctrl-D)',
                               'method': self._quitCmd,
            },
            'echo':          { 'desc': 'echo parameters after expansion',
                               'usage': 'echo <args>',
                               'method': self._echoCmd,
            },
            'set':           { 'desc': 'set attribute, e.g. cachedir',
                               'usage': 'set <attr> <values>',
                               'method': self._setCmd,
            },
            'clear':         { 'desc': 'clear attribute, e.g. print-options',
                               'usage': 'clear <attr>',
                               'method': self._clearCmd,
            },
            'ls-attrs':      { 'desc': 'list attributes',
                               'usage': 'ls-attrs',
                               'method': self._lsAttrsCmd,
            },
            'ls':            { 'desc': 'list filesets',
                               'usage': 'ls',
                               'method': self._lsCmd,
            },
            'ls-caches':     { 'desc': 'list caches',
                               'usage': 'ls-caches',
                               'method': self._lsCachesCmd,
            },
            'fileset':       { 'desc': 'define a fileset',
                               'usage': 'fileset <name> find.gnu.out|find|filter|union <spec>',
                               'method': self._filesetCmd,
            },
            'info':          { 'desc': 'show summary information for a fileset',
                               'usage': 'info [-u|-d|-e|-s] <fileset> [<filter-params>]',
                               'method': self._infoCmd,
            },
            'print':         { 'desc': 'print files in a fileset, optionally filtered, via $PAGER',
                               'usage': 'print <fileset> [<filter-params>] [-by-size] [-depth <depth>]',
                               'method': self._printCmd,
            },
            'delete':        { 'desc': 'delete all files in a fileset, optionally filtered',
                               'usage': 'delete <fileset> [<filter-params>]',
                               'method': self._deleteCmd,
            },
            'update-cache':  { 'desc': 'update all or named caches, by rescanning source filelists',
                               'usage': 'update-cache [<fileset> ...]',
                               'method': self._updateCacheCmd,
            },
            'send-emails':   { 'privileged': True,
                               'desc': 'send emails to owners in fileset, using named template',
                               'usage': 'send-emails <fileset> <email-template> [<override-recipient>]',
                               'method': self._sendEmailsCmd,
            },
        }
        initialize(args)

    def _cached(self, name, fileset):
        if 'cachedir' not in self._attrs:
            raise CLIError("missing attr cachedir")
        cachedirs = self._attrs['cachedir']
        if len(cachedirs) != 1:
            raise CLIError("botched attr cachedir")
        cachedir = cachedirs[0]
        if 'deltadir' not in self._attrs:
            raise CLIError("missing attr deltadir")
        deltadirs = self._attrs['deltadir']
        if len(deltadirs) != 1:
            raise CLIError("botched attr deltadir")
        deltadir = deltadirs[0]
        cache = Cache(name, fileset,
                      os.path.join(cachedir, name),
                      os.path.join(deltadir, name),
                      self._mapper,
                      self._attrs)
        self._caches[name] = cache
        return cache

    def _expandVars(self, toks):
        """Expand environment variables in tokens."""
        for i in range(len(toks)):
            m = True
            while m:
                m = re.search(r"""\$([a-zA-Z]\w*)""", toks[i])
                if m:
                    name = m.group(1)
                    if name in os.environ:
                        val = os.environ[name]
                    else:
                        val = ""
                    toks[i] = toks[i].replace("$%s" % name, val)
                    #print("matched env var %s, val='%s', token now '%s'" % (name, val, toks[i]))

    def _fileset(self, name):
        if name not in self._filesets:
            raise CLIError("no such fileset %s" % name)
        return self._filesets[name]

    def _cache(self, name):
        if name not in self._caches:
            raise CLIError("no such cache %s" % name)
        return self._caches[name]

    def _process(self, line):
        done = False
        toks = shlex.split(line, comments=True)
        self._expandVars(toks)
        if len(toks) >= 1:
            timeCommand = toks[0] == 'time'
            if timeCommand:
                toks = toks[1:]
            cmdName = toks[0]
            if cmdName in self.commands:
                cmd = self.commands[cmdName]
                method = cmd['method']
                usage = cmd['usage']
                privileged = cmd['privileged'] if 'privileged' in cmd else False
                if privileged and os.geteuid() != 0:
                    raise CLIError("attempt to use privileged command %s" % toks[0])
                if profile():
                    pr = cProfile.Profile()
                    pr.enable()
                if timeCommand:
                    t = unix_time(method, (toks, usage))
                    print("time: real %.2fs, user %.2fs, sys %.2fs" % (t['real'], t['user'], t['sys']))
                else:
                    method(toks, usage)
                if profile():
                    pr.disable()
                    ps = pstats.Stats(pr).sort_stats('cumulative')
                    ps.print_stats()
                done = method == self._quitCmd
            else:
                raise CLIError("unknown command %s, try help" % cmdName)
        return done

    def _handleProcess(self, line):
        done = False
        try:
            done = self._process(line)
        except CLIError as e:
            stderr("ERROR %s\n" % e.msg)
        except KeyboardInterrupt:
            stderr("\n")
        return done

    def startup(self):
        for rc in ["/etc/filebutlerrc",
                   os.path.expanduser("~/.filebutlerrc")]:
            try:
                with open(rc) as f:
                    for line in f:
                        self._handleProcess(line)
            except IOError:
                pass

    def execute(self, cmds):
        for cmd in cmds:
            self._handleProcess(cmd)

    def interact(self):
        done = False
        while not done:
            try:
                line = input("fb: ")
                done = self._handleProcess(line)
            except EOFError:
                print("bye")
                done = True
            except KeyboardInterrupt:
                stderr("^C\n")

    def _helpCmd(self, toks, usage):
        for cmdname in sorted(self.commands.keys()):
            cmd = self.commands[cmdname]
            privileged = cmd['privileged'] if 'privileged' in cmd else False
            if not privileged or os.geteuid() == 0:
                print("%-12s - %s\n               %s\n" % (cmdname, cmd['desc'], cmd['usage']))
        cmdname = 'time'
        cmd = { 'desc': 'time a command',
                'usage': 'time <cmd> <args>' }
        print("%-12s - %s\n               %s\n" % (cmdname, cmd['desc'], cmd['usage']))

    def _quitCmd(self, toks, usage):
        pass

    def _echoCmd(self, toks, usage):
        print(' '.join(toks[1:]))

    def _setCmd(self, toks, usage):
        if len(toks) < 2:
            raise CLIError("usage: %s" % usage)
        name = toks[1]
        values = toks[2:]
        self._attrs[name] = values
        if name == 'dataset':
            if len(values) != 2:
                raise CLIError("botched attr %s" % name)
            self._pathway.setDatasetRegex(values[0], values[1])
        if name == 'ignorepathsfrom':
            if len(values) != 1:
                raise CLIError("botched attr %s" % name)
            self._pathway.setIgnorePathsFrom(values[0])

    def _clearCmd(self, toks, usage):
        if len(toks) != 2:
            raise CLIError("usage: %s" % usage)
        name = toks[1]
        del self._attrs[name]
        if name == 'dataset':
            self._pathway.clearDatasetRegex()

    def _lsAttrsCmd(self, toks, usage):
        if len(toks) != 1:
            raise CLIError("usage: %s" % usage)
        for name in sorted(self._attrs.keys()):
            print("%s = %s" % (name, ' '.join(self._attrs[name])))

    def _lsCmd(self, toks, usage):
        if len(toks) != 1:
            raise CLIError("usage: %s" % usage)
        for name in self._filesetNames:
            print(self._filesets[name].description())

    def _lsCachesCmd(self, toks, usage):
        if len(toks) != 1:
            raise CLIError("usage: %s" % usage)
        for name in self._filesetNames:
            if name in self._caches:
                print(self._filesets[name].description())

    def _filesetCmd(self, toks, usage):
        if len(toks) < 3:
            raise CLIError("usage: %s" % usage)
        name = toks[1]
        type = toks[2]
        if name in self._filesets:
            raise CLIError("duplicate fileset %s" % name)
        if type == "find.gnu.out":
            fileset = self._cached(name, GnuFindOutFileset.parse(self._mapper, self._pathway, name, toks[3:]))
        elif type == "find":
            fileset = self._cached(name, FindFileset.parse(self._mapper, self._pathway, name, toks[3:]))
        elif type == "filter":
            if len(toks) < 4:
                raise CLIError("filter requires fileset, criteria")
            filter, _, _ = parseCommandOptions(self._now, toks[4:], filter=True)
            fileset = FilterFileset(name, self._fileset(toks[3]), filter)
        elif type == "union":
            if len(toks) < 4:
                raise CLIError("union requires at least two filesets")
            filesets = []
            for filesetName in toks[3:]:
                filesets.append(self._fileset(filesetName))
            fileset = UnionFileset(name, filesets)
        else:
            raise CLIError("unknown fileset type %s" % type)
        self._filesets[name] = fileset
        self._filesetNames.append(name)

    def _infoCmd(self, toks, usage):
        if len(toks) < 2:
            raise CLIError("usage: %s" % usage)
        if toks[1] == '-u' or toks[1] == '-d' or toks[1] == '-e' or toks[1] == '-s':
            mode = toks[1][1]
            i_fileset = 2
        else:
            mode = 'a'          # all
            i_fileset = 1
        if len(toks) <= i_fileset:
            raise CLIError("usage: %s" % usage)
        name = toks[i_fileset]
        if len(toks) > i_fileset:
            filter, _, _ = parseCommandOptions(self._now, toks[i_fileset + 1:], filter=True)
        else:
            filter = None
        info = self._fileset(name).info(self._attrs, filter)
        if mode == 'a':
            print(info.fmt_total())
        elif mode == 'u':
            print(info.fmt_users())
        elif mode == 'd':
            print(info.fmt_datasets())
        elif mode == 'e':
            for user, userinfo in info.iterusers():
                if user not in self._aliases:
                    print("%s %s" % (user, str(userinfo)))
        elif mode == 's':
            print(info.fmt_sizes())
        else:
            raise CLIError("usage: %s" % usage)

    def _printCmd(self, toks, usage):
        if len(toks) < 2:
            raise CLIError("usage: %s" % usage)
        name = toks[1]
        fileset = self._fileset(name)
        if 'print-options' in self._attrs:
            printOptions = toks[2:] + self._attrs['print-options']
        else:
            printOptions = toks[2:]
        if printOptions != []:
            filter, sorter, grouper = parseCommandOptions(self._now, printOptions, filter=True, sorter=True, grouper=True)
        else:
            filter, sorter, grouper = None, None, Grouper()

        pager = Pager()
        grouper.setOutput(pager.file)
        try:
            for filespec in fileset.sorted(filter, sorter):
                grouper.write(filespec)
            grouper.flush()
        except IOError as e:
            if e.errno == errno.EPIPE:
                pass
            else:
                raise
        finally:
            pager.close()

    def _deleteCmd(self, toks, usage):
        if len(toks) < 2:
            raise CLIError("usage: %s" % usage)
        name = toks[1]
        fileset = self._fileset(name)
        deleteOptions = toks[2:]
        if deleteOptions != []:
            filter, _, _ = parseCommandOptions(self._now, deleteOptions, filter=True)
        else:
            filter = None
        with DeletionLog(self._attrs) as logf:
            # delete directories after their contents
            dirs = []
            mtimes = {}
            for filespec in fileset.select(filter):
                # preserve mtime for parent directory
                parent = os.path.dirname(filespec.path)
                if parent not in mtimes:
                    mtimes[parent] = os.stat(parent).st_mtime
                if filespec.isdir():
                    dirs.append(filespec)
                else:
                    filespec.delete(logf)
            for filespec in sorted(dirs, key=lambda d: d.path, reverse=True):
                filespec.delete(logf)
        # now reset mtimes of anything that's left
        for path, mtime in mtimes.items():
            try:
                os.utime(path, (mtime, mtime))
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # silently do nothing if there's no directory left
                    pass
                elif e.errno == errno.EACCES or e.errno == errno.EPERM:
                    # silently do nothing if we don't have permission to fix mtime
                    pass
                else:
                    raise
        # finally save deletions lists for all caches
        for name in self._caches:
            self._caches[name].saveDeletions()

    def _updateCacheCmd(self, toks, usage):
        if len(toks) == 1:
            for name in sorted(list(self._caches.keys())):
                #print("updating cache %s" % name)
                self._caches[name].update()
        else:
            for name in toks[1:]:
                self._cache(name).update()

    def _attrsAsStringMap(self):
        s = {}
        for key, values in self._attrs.items():
            s[key] = ' '.join(values)
        return s

    def _sendEmailsCmd(self, toks, usage):
        if len(toks) < 3 or len(toks) > 4:
            raise CLIError("usage: %s" % usage)
        if 'emailfrom' not in self._attrs:
            raise CLIError("missing attr emailfrom")
        if 'templatedir' not in self._attrs:
            raise CLIError("missing attr templatedir")
        templatedirs = self._attrs['templatedir']
        if len(templatedirs) != 1:
            raise CLIError("botched attr templatedir")
        templatedir = templatedirs[0]
        emailonly = self._attrs['emailonly'] if 'emailonly' in self._attrs else None
        name = toks[1]
        template = toks[2]
        override_recipient = None if len(toks) == 3 else toks[3]
        subject_path = os.path.join(templatedir, "%s.subject" % template)
        with open(subject_path, 'r') as f:
            subject_template = string.Template(f.read().replace('\n', '').strip())
        body_path = os.path.join(templatedir, "%s.body" % template)
        with open(body_path, 'r') as f:
            body_template = string.Template(f.read())
        fileset = self._fileset(name)
        m = self._attrsAsStringMap()
        m['fileset'] = name
        m['fileset_descriptor'] = str(fileset.description())
        s = smtplib.SMTP('localhost')
        sender = ' '.join(self._attrs['emailfrom'])
        for user, userfileinfo in fileset.info(self._attrs).iterusers():
            if user in self._aliases and (emailonly is None or user in emailonly):
                recipient = self._aliases[user] if override_recipient is None else override_recipient
                user_fileset = FilterFileset("%s-%s" % (name, user), fileset, Filter(owner=user))
                user_info = user_fileset.info(self._attrs)
                m['username'] = user
                m['info'] = user_info.fmt_total()
                m['info_datasets'] = user_info.fmt_datasets()
                #print("==================== %s %s\n\n%s" % (user_email, subject, body))
                msg = email.mime.text.MIMEText(body_template.substitute(m))
                msg['Subject'] = subject_template.substitute(m)
                msg['From'] = sender
                msg['To'] = recipient
                s.sendmail(sender, [recipient], msg.as_string())
        s.quit()



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import os
import subprocess

class Pager(object):

    def __init__(self):
        if 'PAGER' in os.environ:
            pager = os.environ['PAGER']
        else:
            pager = 'less'
        self._pager = subprocess.Popen(pager, stdin=subprocess.PIPE, universal_newlines=True)
        self.file = self._pager.stdin

    def close(self, force=False):
        self.file.close()
        if force:
            self._pager.terminate()
        self._pager.wait()



# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
class FilesetSelector(object):

    def __init__(self, owner=None, dataset=None, sizebucket=None):
        self.owner = owner
        self.dataset = dataset
        self.sizebucket = sizebucket

    def withOwner(self, owner):
        return self.__class__(owner, self.dataset, self.sizebucket)

    def withDataset(self, dataset):
        return self.__class__(self.owner, dataset, self.sizebucket)

    def withSizebucket(self, sizebucket):
        return self.__class__(self.owner, self.dataset, sizebucket)



#!/usr/bin/env python
#
# distutils setup script for filebutler package

from setuptools import setup, find_packages

long_description = """Filebutler is a utility for managing old files in huge directory trees.  Huge means of the order of 100 million files.  The motivation is that find is far too slow on directory trees with this many files.  Even working with filelists generated offline by find can be rather slow for interactive use.  Filebutler improves on this by structuring filelists by age, by size, by owner, and by dataset.
"""

setup(name='filebutler',
      use_scm_version=True,
      setup_requires=['setuptools_scm', 'setuptools_scm_git_archive'],
      description='Utility for managing old files in large directory structures',
      long_description=long_description,
      author='Simon Guest',
      author_email='simon.guest@tesujimath.org',
      url='https://github.com/tesujimath/filebutler',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: POSIX :: Linux',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: System :: Systems Administration',
          'Topic :: Utilities',
      ],
      packages=find_packages(),
      entry_points={
        'console_scripts': [
            'filebutler = filebutler.__main__:main',
        ],
      },
      package_data={
          'doc': ['*.rst'],
          'examples': ['*'],
      },
      license='GPLv3',
      install_requires=[
          'future',
          'pytz',
          'setuptools',
          'tzlocal',
      ],
      python_requires='>=2.7',
     )



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import range
from builtins import object
import datetime
import os.path
import re
import shutil

from .Buckets import Buckets
from .Filter import Filter
from .FilespecMerger import FilespecMerger
from .PooledFile import listdir
from .util import str2size, verbose_stderr, debug_log

class SizeFilesetCache(object):

    def __init__(self, path, deltadir, mapper, attrs, sel, next):
        self._path = path
        self._deltadir = deltadir
        self._mapper = mapper
        self._attrs = attrs
        self._sel = sel
        self._next = next
        self._sizebuckets = Buckets([str2size(s) for s in self._attrs['sizebuckets']] if 'sizebuckets' in self._attrs else [])
        self._filesets = [None] * self._sizebuckets.len

    def _subpath(self, b):
        return os.path.join(self._path, str(b))

    def _subdeltadir(self, b):
        return os.path.join(self._deltadir, str(b))

    def _fileset(self, i):
        """On demand creation of child filesets."""
        b = self._sizebuckets.bound(i)
        fileset = self._filesets[i]
        if fileset is None:
            fileset = self._next(self._subpath(b), self._subdeltadir(b), self._mapper, self._attrs, self._sel.withSizebucket(b))
            self._filesets[i] = fileset
        return fileset

    def select(self, filter=None):
        merger = FilespecMerger()
        for i in range(self._sizebuckets.len):
            minSize, maxSize = self._sizebuckets.minmax(i)
            if filter is None or filter.sizeGeq is None or maxSize is None or maxSize >= filter.sizeGeq:
                if filter is not None and filter.sizeGeq is not None and minSize >= filter.sizeGeq:
                    f1 = Filter.clearSize(filter)
                else:
                    f1 = filter
                merger.add(self._fileset(i).select(f1))
        # no yield from in python 2, so:
        for filespec in merger.merge():
            yield filespec

    def merge_info(self, acc, filter=None):
        #debug_log("SizeFilesetCache(%s) merge_info\n" % self._path)

        for i in range(self._sizebuckets.len):
            minSize, maxSize = self._sizebuckets.minmax(i)
            if filter is None or filter.sizeGeq is None or maxSize is None or maxSize >= filter.sizeGeq:
                if filter is not None and filter.sizeGeq is not None and minSize >= filter.sizeGeq:
                    f1 = Filter.clearSize(filter)
                else:
                    f1 = filter
                self._fileset(i).merge_info(acc, f1)

    def add(self, filespec):
        fileset = self._fileset(self._sizebuckets.indexContaining(filespec.size))
        fileset.add(filespec)

    def finalize(self):
        for fs in self._filesets:
            if fs is not None:
                fs.finalize()

    def saveDeletions(self):
        #debug_log("SizeFilesetCache(%s)::saveDeletions\n" % self._path)
        for fs in self._filesets:
            if fs is not None:
                fs.saveDeletions()



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
import errno
import os

from .FatalError import FatalError
from .util import debug_log

def listdir(path):
    """Just like os.listdir, but close pooled files if hit too many open."""
    try:
        files = os.listdir(path)
    except OSError as e:
        if e.errno == errno.EMFILE:
            debug_log("listdir failed for %s, flush all pooled files\n" % path)
            PooledFile.flushAll()
            files = os.listdir(path)
        else:
            raise
    return files

class PooledFile(object):
    """A File object in a pool, which catches too many open files."""

    pool = {}

    @classmethod
    def flushAll(cls):
        """Flush all files in the pool."""
        for f in list(cls.pool.keys()):
            f.flush()

    def __init__(self, name, mode='r'):
        self._name = name
        self._mode = mode
        self._file = None
        self._readpos = None
        self.__class__.pool[self] = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _readmode(self):
        return self._mode == 'r'

    def _open(self):
        try:
            self._file = open(self._name, self._mode)
        except IOError as e:
            if e.errno == errno.EMFILE:
                debug_log("open failed for %s, flush all pooled files\n" % self._name)
                # It would be better to implement an LRU cache, but that
                # hardly seems worth it.  So simply close them all.
                self.__class__.flushAll()
                self._file = open(self._name, self._mode)
            else:
                raise
        if self._readmode() and self._readpos is not None:
            self._file.seek(self._readpos)

    def __iter__(self):
        """Iterate over the lines in a file opened for read."""
        done = False
        while not done:
            if self._file is None:
                self._open()
            line = self._file.readline()
            if line != '':
                yield line
            else:
                done = True

    def seek(self, offset):
        if self._file is not None:
            self._file.seek(offset)
        elif self._readmode():
            self._readpos = offset
        else:
            raise FatalError("seek on closed PooledFile for write, unsupported")

    def tell(self):
        if self._file is not None:
            return self._file.tell()
        elif self._readmode():
            return self._readpos
        else:
            raise FatalError("tell on closed PooledFile for write, unsupported")

    def write(self, s):
        if self._file is None:
            self._open()
        self._file.write(s)

    def flush(self):
        if self._file is not None:
            if self._readmode():
                self._readpos = self._file.tell()
            else:
                # subsequently need to append
                self._mode = 'a'
            self._file.close()
            self._file = None

    def close(self):
        if self._file is not None:
            self._file.close()
            self._file = None
        del self.__class__.pool[self]



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import object
from copy import copy

from .FilesetInfo import FilesetInfo
from .Filespec import Filespec

class Grouper(object):

    def __init__(self, collapse=None):
        self._collapse = collapse
        self._collapsing = None

    def setOutput(self, f):
        self._f = f
        self._width = 0

    def _collapsed(self, path):
        n = self._collapse
        slash = path.find('/')
        while slash >= 0 and n > 0:
            slash = path.find('/', slash + 1)
            n -= 1
        if slash >= 0:
            return path[:slash]
        else:
            return path

    def _writeFilespec(self, filespec):
        s, self._width = filespec.format(self._width)
        self._f.write("%s\n" % s)

    def write(self, filespec):
        if self._collapse is None:
            self._writeFilespec(filespec)
        else:
            p = self._collapsed(filespec.path)
            if self._collapsing is None or p != self._collapsing.path:
                if self._collapsing is not None:
                    self._writeFilespec(self._collapsing)
                self._collapsing = copy(filespec)
                self._collapsing.path = p
                self._collapsing.perms = 'd' + self._collapsing.perms[1:]
            self._collapsing.size += filespec.size

    def flush(self):
        if self._collapse is not None and self._collapsing is not None:
            self._writeFilespec(self._collapsing)



from __future__ import absolute_import
# Copyright 2017 Simon Guest
#
# This file is part of filebutler.
#
# Filebutler is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Filebutler is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with filebutler.  If not, see <http://www.gnu.org/licenses/>.

from builtins import str
from builtins import range
from builtins import object

class Buckets(object):

    def __init__(self, bounds):
        self._bounds = [0]
        for b in bounds:
            self._insert(b)

    def _insert(self, b):
        for i in range(len(self._bounds)):
            if b == self._bounds[i]:
                # already present, do nothing
                return
            elif b < self._bounds[i]:
                self._bounds.insert(i, b)
                return
        self._bounds.append(b)

    def indexContaining(self, x):
        """Return slot containing a given value."""
        for i in reversed(list(range(len(self._bounds)))):
            if x >= self._bounds[i]:
                return i
        raise ValueError("no slot containing %d, %s" % (x, str(self._bounds)))

    def index(self, b):
        """Return slot for given bound."""
        for i in reversed(list(range(len(self._bounds)))):
            if b == self._bounds[i]:
                return i
        raise ValueError("no slot with bound %d, %s" % (b, str(self._bounds)))

    def bound(self, i):
        return self._bounds[i]

    def minmax(self, i):
        """Return (min, max), or None for max if this is the last bucket."""
        return self._bounds[i], (self._bounds[i + 1] - 1 if i < len(self._bounds) - 1 else None)

    @property
    def len(self):
        return len(self._bounds)
